# coding=utf-8
from __future__ import absolute_import

import numpy as np
from rasterio import transform
from rasterio.io import MemoryFile

from .. import _nodata, get_resolution_in_meters

CONTENT_TYPE = "image/tiff"


def GeoTIFF():
    def _format((data, (data_bounds, data_crs)), data_format):
        if data_format is not "raw":
            raise Exception("raw data is required")

        (count, height, width) = data.shape

        if count == 1:
            resolution = get_resolution_in_meters((data_bounds, data_crs),
                                                  (height, width))

            # downsample to int16 if ground resolution is more than 10 meters
            # (at the equator)
            if resolution[0] > 10 and resolution[1] > 10:
                data = data.astype(np.int16)
                data.fill_value = _nodata(data.dtype)

        if np.issubdtype(data.dtype, np.float):
            predictor = 3
        else:
            predictor = 2

        meta = {
            "blockxsize": 512 if width >= 512 else width,
            "blockysize": 512 if height >= 512 else height,
            "compress": "deflate",
            "count": count,
            "crs": data_crs,
            "dtype": data.dtype,
            "driver": "GTiff",
            "nodata": data.fill_value if data.dtype != np.uint8 else None,
            "predictor": predictor,
            "height": height,
            "width": width,
            "tiled": width >= 512 and height >= 512,
            "transform": transform.from_bounds(
                *data_bounds,
                width=width,
                height=height),
        }

        with MemoryFile() as memfile:
            with memfile.open(**meta) as dataset:
                # TODO not true for aerial imagery
                dataset.update_tags(AREA_OR_POINT="Point")
                dataset.write(data.filled())

            return (CONTENT_TYPE, memfile.read())

    return _format



# coding=utf-8
from __future__ import absolute_import

import logging
import os
import urlparse

from rasterio import warp

from marblecutter import get_zoom
from psycopg2.pool import ThreadedConnectionPool

from . import WGS84_CRS, Catalog

Infinity = float("inf")


class PostGISCatalog(Catalog):
    def __init__(self, table="footprints", database_url=os.getenv("DATABASE_URL"), geometry_column="wkb_geometry"):
        if database_url is None:
            raise Exception(
                "Database URL must be provided, either as an arg or as DATABASE_URL."
            )
        urlparse.uses_netloc.append('postgis')
        urlparse.uses_netloc.append('postgres')
        url = urlparse.urlparse(database_url)

        self._pool = ThreadedConnectionPool(
            1,
            16,
            database=url.path[1:],
            user=url.username,
            password=url.password,
            host=url.hostname,
            port=url.port, )

        self._log = logging.getLogger(__name__)
        self.table = table
        self.geometry_column = geometry_column

    def get_sources(self, (bounds, bounds_crs), resolution):
        zoom = get_zoom(max(resolution))

        self._log.info("Resolution: %s; equivalent zoom: %d", resolution, zoom)

        # TODO get sources in native CRS of the target
        query = """
            SELECT
                url,
                source,
                resolution
            FROM (
                SELECT
                    DISTINCT ON (url) url,
                    source,
                    resolution,
                    priority,
                    -- group sources by approximate resolution
                    round(resolution) rounded_resolution,
                    -- measure the distance from centroids to prioritize overlap
                    ST_Centroid({geometry_column}) <-> ST_Centroid(
                        ST_SetSRID(
                            'BOX(%(minx)s %(miny)s, %(maxx)s %(maxy)s)'::box2d,
                            4326)) distance
                FROM {table}
                WHERE {geometry_column} && ST_SetSRID(
                    'BOX(%(minx)s %(miny)s, %(maxx)s %(maxy)s)'::box2d, 4326)
                    AND %(zoom)s BETWEEN min_zoom AND max_zoom
                    AND enabled = true
                ORDER BY url
            ) AS _
            ORDER BY priority ASC, rounded_resolution ASC, distance ASC
        """.format(table=self.table, geometry_column=self.geometry_column)

        # height and width of the CRS
        ((left, right), (bottom, top)) = warp.transform(
            bounds_crs, WGS84_CRS, bounds[::2], bounds[1::2])

        conn = self._pool.getconn()
        try:
            with conn.cursor() as cur:
                cur.execute(query, {
                    "minx": left if left != Infinity else -180,
                    "miny": bottom if bottom != Infinity else -90,
                    "maxx": right if right != Infinity else 180,
                    "maxy": top if top != Infinity else 90,
                    "zoom": zoom,
                })

                return cur.fetchall()
        finally:
            self._pool.putconn(conn)



# noqa
# coding=utf-8
from __future__ import absolute_import, print_function

from StringIO import StringIO

import matplotlib

# pick a matplotlib backend to pre-empt loading addition modules
matplotlib.use("Agg")

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap


CONTENT_TYPES = {
    "png": "image/png",
}

GREY_HILLS_RAMP = {
    "red": [(0.0, 0.0, 0.0), (0.25, 0.0, 0.0), (180 / 255.0, 0.5, 0.5),
            (1.0, 170 / 255.0, 170 / 255.0)],
    "green": [(0.0, 0.0, 0.0), (0.25, 0.0, 0.0), (180 / 255.0, 0.5, 0.5),
              (1.0, 170 / 255.0, 170 / 255.0)],
    "blue": [(0.0, 0.0, 0.0), (0.25, 0.0, 0.0), (180 / 255.0, 0.5, 0.5),
             (1.0, 170 / 255.0, 170 / 255.0)],
}

GREY_HILLS = LinearSegmentedColormap("grey_hills", GREY_HILLS_RAMP)


def ColorRamp(output_format="png", colormap=GREY_HILLS):
    def _format((data, (data_bounds, data_crs)), data_format):
        if data_format is not "raw":
            raise Exception("raw data is required")

        if data.dtype != np.uint8:
            raise Exception("data must be uint8")

        out = StringIO()
        plt.imsave(
            out,
            data[0],
            cmap=colormap,
            vmin=0,
            vmax=255,
            format=output_format)

        return (CONTENT_TYPES[output_format], out.getvalue())

    return _format



# noqa
# coding=utf-8
from __future__ import absolute_import

import mercantile
from affine import Affine
from rasterio.crs import CRS

from . import render

TILE_SHAPE = (256, 256)
WEB_MERCATOR_CRS = CRS.from_epsg(3857)


def render_tile(tile,
                sources,
                transformation=None,
                format=None,
                scale=1):
    """Render a tile into Web Mercator."""
    bounds = mercantile.xy_bounds(tile)

    return render(
        (bounds, WEB_MERCATOR_CRS),
        sources,
        map(int, Affine.scale(scale) * TILE_SHAPE),
        WEB_MERCATOR_CRS,
        format=format,
        transformation=transformation)



# coding=utf-8
from __future__ import absolute_import

import numpy as np

from rasterio import warp
from rasterio.crs import CRS

from .. import crop, get_extent, get_resolution

WGS84_CRS = CRS.from_epsg(4326)


class TransformationBase:
    buffer = 0

    def __init__(self, collar=0):
        self.collar = collar

    def expand(self, (bounds, bounds_crs), shape):
        buffer = self.buffer
        collar = self.collar
        effective_buffer = buffer + collar

        resolution = get_resolution((bounds, bounds_crs), shape)

        # apply buffer
        bounds_orig = bounds
        shape = [dim + (2 * effective_buffer) for dim in shape]
        bounds = [
            p - (buffer * resolution[i % 2])
            if i < 2 else p + (effective_buffer * resolution[i % 2])
            for i, p in enumerate(bounds)
        ]

        #
        left = right = bottom = top = buffer

        # adjust bounds + shape if bounds extends outside the extent
        extent = get_extent(bounds_crs)

        # TODO this is all or nothing right now
        # also: float precision
        if bounds[0] < extent[0]:
            shape[1] -= effective_buffer
            bounds[0] = bounds_orig[0]
            left = 0

        if bounds[2] > extent[2]:
            shape[1] -= effective_buffer
            bounds[2] = bounds_orig[2]
            right = 0

        if bounds[1] < extent[1]:
            shape[0] -= effective_buffer
            bounds[1] = bounds_orig[1]
            bottom = 0

        if bounds[3] > extent[3]:
            shape[0] -= effective_buffer
            bounds[3] = bounds_orig[3]
            top = 0

        return (bounds, bounds_crs), shape, (left, bottom, right, top)

    def postprocess(self, (data, (bounds, data_crs)), data_format, offsets):
        (data, (cropped_bounds, data_crs)) = crop((data, (bounds, data_crs)),
                                                  data_format, offsets)

        if self.collar > 0:
            extent = get_extent(data_crs)

            if data_format == "RGBA":
                if np.isclose(bounds[0], extent[0]):
                    # wrap left
                    cols = data[:, :self.collar]
                    data = np.hstack((cols, data))

                if np.isclose(bounds[2], extent[2]):
                    # wrap right
                    cols = data[:, -self.collar:]
                    data = np.hstack((data, cols))

                if np.isclose(bounds[1], extent[1]):
                    # repeat bottom
                    rows = data[-self.collar:]
                    data = np.vstack((data, rows))

                if np.isclose(bounds[3], extent[3]):
                    # repeat top
                    rows = data[:self.collar]
                    data = np.vstack((rows, data))
            else:
                raise Exception(
                    "Unsupported format for collar postprocessing: {}".format(
                        data_format))

        return (data, (cropped_bounds, data_crs))

    def transform(self, (data, (bounds, crs))):
        return (data, "raw")


def apply_latitude_adjustments(data, (bounds, crs)):
    (_, height, width) = data.shape

    ys = np.interp(np.arange(height), [0, height - 1], [bounds[3], bounds[1]])
    xs = np.empty_like(ys)
    xs.fill(bounds[0])

    longitudes, latitudes = warp.transform(crs, WGS84_CRS, xs, ys)

    factors = 1 / np.cos(np.radians(latitudes))

    # convert to 2d array, rotate 270º, scale data
    return data * np.rot90(np.atleast_2d(factors), 3)



from rasterio.crs import CRS

WGS84_CRS = CRS.from_epsg(4326)


class Catalog(object):
    @property
    def bounds(self):
        return [-180, -85.05113, 180, 85.05113]

    @property
    def center(self):
        return [0, 0, 2]

    @property
    def id(self):
        return None

    @property
    def maxzoom(self):
        return 22

    @property
    def metadata_url(self):
        return None

    @property
    def minzoom(self):
        return 0

    @property
    def name(self):
        return "Untitled"

    @property
    def provider(self):
        return None

    @property
    def provider_url(self):
        return None

    def get_sources(self, (bounds, bounds_crs), resolution):
        raise NotImplemented



# noqa
# coding=utf-8
from __future__ import absolute_import, division, print_function

import numpy as np

from .utils import TransformationBase


class Image(TransformationBase):
    def transform(self, (data, (bounds, crs))):
        (count, height, width) = data.shape

        if 3 > count > 4:
            raise Exception("Source data must be 3 or 4 bands")

        if count == 4:
            raise Exception(
                "Variable opacity (alpha channel) not yet implemented")

        rgb = np.ma.transpose(data.astype(np.uint8), [1, 2, 0])
        if data.mask.any():
            a = np.logical_and.reduce(~data.mask).astype(np.uint8) * 255
        else:
            a = np.full((rgb.shape[:-1]), 255, np.uint8)

            # Nearblack filtering for collar removal--partial, as edge values
            # will have been resampled in such a way that they don't retain
            # their crispness.
            # See https://stackoverflow.com/a/22631583 for neighborhood
            # filtering
            # sums = np.add.reduce(data)
            # threshold = 64
            # a = np.logical_and(sums > threshold, sums <
            #                    (255 * 3) - threshold).astype(np.uint8) * 255

        return (np.dstack((rgb, a)), 'RGBA')



# coding=utf-8
from __future__ import absolute_import, division, print_function

import logging
import math
import os
import warnings

from haversine import haversine
import numpy as np
import rasterio
from rasterio import transform, warp, windows
from rasterio._err import CPLE_OutOfMemoryError
from rasterio.crs import CRS
from rasterio.transform import Affine
from rasterio.vrt import WarpedVRT
from rasterio.warp import Resampling
from rasterio.windows import Window
from scipy import ndimage

from . import mosaic
from .stats import Timer

WEB_MERCATOR_CRS = CRS.from_epsg(3857)
WGS84_CRS = CRS.from_epsg(4326)
LOG = logging.getLogger(__name__)

EXTENTS = {
    str(WEB_MERCATOR_CRS): (-20037508.342789244, -20037508.342789244,
                            20037508.342789244, 20037508.342789244),
}


class NoDataAvailable(Exception):
    pass


def _isimage(data_format):
    return data_format.upper() in ["RGB", "RGBA"]


def _mask(data, nodata):
    if np.issubdtype(data.dtype, float):
        return np.ma.masked_values(data, nodata, copy=False)

    return np.ma.masked_equal(data, nodata, copy=False)


def _nodata(dtype):
    if np.issubdtype(dtype, int):
        return np.iinfo(dtype).min
    else:
        return np.finfo(dtype).min


def crop((data, (bounds, data_crs)), data_format, offsets):
    left, bottom, right, top = offsets

    if _isimage(data_format):
        width, height, _ = data.shape
        t = transform.from_bounds(*bounds, width=width, height=height)

        data = data[top:height - bottom, left:width - right, :]

        cropped_window = windows.Window(left, top, width, height)
        cropped_bounds = windows.bounds(cropped_window, t)

        return (data, (cropped_bounds, data_crs))

    _, height, width = data.shape
    t = transform.from_bounds(*bounds, width=width, height=height)

    data = data[:, top:height - bottom, left:width - right]

    cropped_window = windows.Window(left, top, width, height)
    cropped_bounds = windows.bounds(cropped_window, t)

    return (data, (cropped_bounds, data_crs))


def get_extent(crs):
    return EXTENTS[str(crs)]


def get_resolution((bounds, crs), (height, width)):
    t = transform.from_bounds(*bounds, width=width, height=height)

    return abs(t.a), abs(t.e)


def get_resolution_in_meters((bounds, crs), (height, width)):
    if crs.is_geographic:
        left = (bounds[0], (bounds[1] + bounds[3]) / 2)
        right = (bounds[2], (bounds[1] + bounds[3]) / 2)
        top = ((bounds[0] + bounds[2]) / 2, bounds[3])
        bottom = ((bounds[0] + bounds[2]) / 2, bounds[1])

        return (haversine(left, right) * 1000 / width,
                haversine(top, bottom) * 1000 / height)

    return get_resolution((bounds, crs), (height, width))


def get_source(path):
    """Cached source opening."""
    with rasterio.Env(CPL_VSIL_CURL_ALLOWED_EXTENSIONS=".vrt,.tif,.ovr,.msk"):
        return rasterio.open(path)


def get_zoom(resolution, op=round):
    return int(
        op(
            math.log((2 * math.pi * 6378137) / (resolution * 256)) / math.log(
                2)))


def read_window(src, (bounds, bounds_crs), (height, width)):
    target_shape = (height, width)
    # TODO use this for DEMs (not all single-band sources) to avoid stairstepping artifacts
    if src.count == 1 and bounds_crs == WEB_MERCATOR_CRS:
        # special case for web Mercator; use a target image size that most
        # closely matches the source resolution (and is a power of 2)
        zoom = min(22,
                   get_zoom(
                       max(
                           get_resolution_in_meters((src.bounds, src.crs), (
                               src.height, src.width))),
                       op=math.ceil))

        dst_width = dst_height = (2**zoom) * 256
        extent = get_extent(bounds_crs)
        resolution = ((extent[2] - extent[0]) / dst_width,
                      (extent[3] - extent[1]) / dst_height)

        dst_transform = Affine(resolution[0], 0.0, extent[0], 0.0,
                               -resolution[1], extent[3])
    else:
        # use a target image size that most closely matches the target
        # resolution
        # calculate natural width, height, and transform (so the mask can be
        # warped)
        # TODO providing resolution reduces the target dimensions but loses the
        # ability to read overviews

        # if raster is overly-large, approximate the transform based on
        # a scaled-down version and scale it back after
        attempts = 0
        scale_factor = 1
        dst_transform = None

        while (dst_transform is None and src.width // scale_factor > 0
               and src.height // scale_factor > 0):
            try:
                (dst_transform, dst_width,
                 dst_height) = warp.calculate_default_transform(
                     src.crs, bounds_crs, src.width // scale_factor,
                     src.height // scale_factor, *src.bounds)

                scale = Affine.scale(scale_factor, scale_factor)

                dst_transform *= ~scale
                dst_width, dst_height = scale * (dst_width, dst_height)
            except (MemoryError, CPLE_OutOfMemoryError):
                attempts += 1
                scale_factor = 2 * attempts

    # Some OAM sources have invalid NODATA values (-1000 for a file with a
    # dtype of Byte). rasterio returns None under these circumstances
    # (indistinguishable from sources that actually have no NODATA values).
    # Providing a synthetic value "correctly" masks the output at the expense
    # of masking valid pixels with that value. This was previously (partially;
    # in the form of the bounding box but not NODATA pixels) addressed by
    # creating a VRT that mapped the mask to an alpha channel (something we
    # can't do w/o adding nDstAlphaBand to rasterio/_warp.pyx).
    #
    # Creating external masks and reading them separately (as below) is a
    # better solution, particularly as it avoids artifacts introduced when the
    # NODATA values are resampled using something other than nearest neighbor.

    # TODO resampling alg should be a catalog property
    with WarpedVRT(
            src,
            src_nodata=src.nodata,
            dst_crs=bounds_crs,
            dst_width=dst_width,
            dst_height=dst_height,
            dst_transform=dst_transform,
            resampling=Resampling.lanczos) as vrt:
        window = vrt.window(*bounds)
        dst_window = window.crop(vrt.height, vrt.width)

        if round(dst_window.width) == 0 or round(dst_window.height) == 0:
            return None

        resolution = get_resolution((bounds, bounds_crs), target_shape)
        src_resolution = get_resolution((vrt.bounds, vrt.crs), vrt.shape)
        scale_factor = (round(window.width / target_shape[1], 6), round(
            window.height / target_shape[0], 6))

        extends_up = False
        extends_left = False

        if bounds[0] < vrt.bounds[0]:
            width = int(
                math.ceil(target_shape[1] * (dst_window.width / window.width)))
            extends_left = True

        if bounds[1] < vrt.bounds[1]:
            height = int(
                math.ceil(target_shape[0] * (dst_window.height / window.height
                                             )))

        if bounds[2] > vrt.bounds[2]:
            width = int(
                math.ceil(target_shape[1] * (dst_window.width / window.width)))

        if bounds[3] > vrt.bounds[3]:
            height = int(
                math.ceil(target_shape[0] * (dst_window.height / window.height
                                             )))
            extends_up = True

        if vrt.count == 1 and (
                scale_factor[0] < 1 or scale_factor[1] < 1
                or src_resolution[0] > resolution[0]
                or src_resolution[1] > resolution[1]
        ) and round(dst_window.width) > 1.0 and round(dst_window.height) > 1.0:
            # scale_factor will always be (1.0, 1.0) unless using the web
            # Mercator-specific calculations
            # instead, compare source resolution (in m) to resolution (as
            # scale_factor) and modify dst_window to correspond to a smaller,
            # lower resolution window (as target_window)

            # TODO sources like ETOPO1 may end up with funky y scale factors
            # (for web Mercator, due to distortion)
            # if these problems can be addressed, the web Mercator-specific
            # calculations (above, for single-band sources) can be removed
            if scale_factor[0] < 1 or scale_factor[1] < 1:
                scaled_transform = vrt.transform * Affine.scale(*scale_factor)
                target_window = windows.from_bounds(
                    *bounds, transform=scaled_transform)
            else:
                scale_factor = (resolution[0] / src_resolution[0],
                                resolution[1] / src_resolution[1])
                target_window = Window(dst_window.col_off * scale_factor[0],
                                       dst_window.row_off * scale_factor[1],
                                       dst_window.width * scale_factor[0],
                                       dst_window.height * scale_factor[1])

            # buffer apparently needs to be 50% of the target size in order
            # for spline knots to match between adjacent tiles
            # however, to avoid creating overly-large uncropped areas, we
            # limit the buffer size to 2048px on a side
            # TODO the resulting window is still much too large (e.g.
            # 18/151153/84343@2x)
            buffer_pixels = (min(target_window.width / 2,
                                 math.ceil(2048 * scale_factor[0])), min(
                                     target_window.height / 2,
                                     math.ceil(2048 * scale_factor[1])))

            r, c = dst_window.toranges()
            window = Window.from_slices((max(
                0, r[0] - buffer_pixels[1]), r[1] + buffer_pixels[1]), (max(
                    0, c[0] - buffer_pixels[0]), c[1] + buffer_pixels[0]))

            data = vrt.read(window=window)

            # mask with NODATA values
            if vrt.nodata is not None:
                data = _mask(data, vrt.nodata)
            else:
                data = np.ma.masked_array(data, mask=False)

            mask = data.mask

            order = 3

            if mask.any():
                # need to preserve NODATA; drop spline interpolation order to 1
                order = 1

            LOG.info(
                "Applying spline interpolation with order %d (scale factor: %s)",
                order, scale_factor)

            zoom = (1, round(1 / scale_factor[1]), round(1 / scale_factor[0]))

            LOG.info("target dimensions: %s", (data.shape[1] * zoom[1],
                                               data.shape[2] * zoom[2]))

            # resample data, respecting NODATA values
            data = ndimage.zoom(
                # prevent resulting values from producing cliffs
                data.astype(np.float32),
                zoom,
                order=order)

            scaled_buffer = (int((data.shape[2] - width) / 2), int(
                (data.shape[1] - height) / 2))

            # crop data
            data = data[:, scaled_buffer[1]:scaled_buffer[1] + height,
                        scaled_buffer[0]:scaled_buffer[0] + width]

            if len(mask.shape) > 0:
                mask = ndimage.zoom(mask, zoom, mode='nearest')

                # crop mask
                mask = mask[:, scaled_buffer[1]:scaled_buffer[1] + height,
                            scaled_buffer[0]:scaled_buffer[0] + width]

            # copy the mask over
            data = np.ma.masked_array(data, mask=mask)
        else:
            data = vrt.read(
                out_shape=(vrt.count, height, width), window=dst_window)

            # mask with NODATA values
            if vrt.nodata is not None:
                data = _mask(data, vrt.nodata)
            else:
                data = np.ma.masked_array(data, mask=False)

        # expand to match the target shape
        # up / down offset
        if extends_up:
            data = np.ma.concatenate((np.ma.masked_array(
                np.zeros(
                    (vrt.count, target_shape[0] - data.shape[1],
                     target_shape[1] - (target_shape[1] - data.shape[2])),
                    dtype=data.dtype),
                mask=True), data), 1)
        else:
            data = np.ma.concatenate((data, np.ma.masked_array(
                np.zeros(
                    (vrt.count, target_shape[0] - data.shape[1],
                     target_shape[1] - (target_shape[1] - data.shape[2])),
                    dtype=data.dtype),
                mask=True)), 1)

        # left / right offset
        if extends_left:
            data = np.ma.concatenate((np.ma.masked_array(
                np.zeros(
                    (vrt.count,
                     target_shape[0] - (target_shape[0] - data.shape[1]),
                     target_shape[1] - data.shape[2]),
                    dtype=data.dtype),
                mask=True), data), 2)
        else:
            data = np.ma.concatenate((data, np.ma.masked_array(
                np.zeros(
                    (vrt.count,
                     target_shape[0] - (target_shape[0] - data.shape[1]),
                     target_shape[1] - data.shape[2]),
                    dtype=data.dtype),
                mask=True)), 2)

        data = data.astype(np.float32)

    # open the mask separately so we can take advantage of its overviews
    try:
        warnings.simplefilter("ignore")
        with rasterio.open("{}.msk".format(src.name), crs=src.crs) as mask_src:
            with WarpedVRT(
                    mask_src,
                    src_crs=src.crs,
                    src_transform=src.transform,
                    dst_crs=bounds_crs,
                    dst_width=dst_width,
                    dst_height=dst_height,
                    dst_transform=dst_transform) as mask_vrt:
                warnings.simplefilter("default")
                window = mask_vrt.window(*bounds)
                dst_window = window.crop(mask_vrt.height, mask_vrt.width)

                if round(dst_window.width) == 0 or round(
                        dst_window.height) == 0:
                    raise Exception("No mask")

                extends_up = False
                extends_left = False

                if bounds[0] < vrt.bounds[0]:
                    width = int(
                        math.ceil(target_shape[1] * (
                            dst_window.width / window.width)))
                    extends_left = True

                if bounds[1] < vrt.bounds[1]:
                    height = int(
                        math.ceil(target_shape[0] * (
                            dst_window.height / window.height)))

                if bounds[2] > vrt.bounds[2]:
                    width = int(
                        math.ceil(target_shape[1] * (
                            dst_window.width / window.width)))

                if bounds[3] > vrt.bounds[3]:
                    height = int(
                        math.ceil(target_shape[0] * (
                            dst_window.height / window.height)))
                    extends_up = True

                mask = mask_vrt.read(
                    out_shape=(mask_vrt.count, height, width),
                    window=dst_window)

                # expand to match the target shape
                # up / down offset
                if extends_up:
                    mask = np.ma.concatenate((np.ma.masked_array(
                        np.zeros(
                            (mask_vrt.count, target_shape[0] - mask.shape[1],
                             target_shape[1] -
                             (target_shape[1] - mask.shape[2])),
                            dtype=mask.dtype),
                        mask=True), mask), 1)
                else:
                    mask = np.ma.concatenate((mask, np.ma.masked_array(
                        np.zeros(
                            (mask_vrt.count, target_shape[0] - mask.shape[1],
                             target_shape[1] -
                             (target_shape[1] - mask.shape[2])),
                            dtype=mask.dtype),
                        mask=True)), 1)

                # left / right offset
                if extends_left:
                    mask = np.ma.concatenate((np.ma.masked_array(
                        np.zeros(
                            (mask_vrt.count, target_shape[0] -
                             (target_shape[0] - mask.shape[1]),
                             target_shape[1] - mask.shape[2]),
                            dtype=mask.dtype),
                        mask=True), mask), 2)
                else:
                    mask = np.ma.concatenate((mask, np.ma.masked_array(
                        np.zeros(
                            (mask_vrt.count, target_shape[0] -
                             (target_shape[0] - mask.shape[1]),
                             target_shape[1] - mask.shape[2]),
                            dtype=mask.dtype),
                        mask=True)), 2)

                data.mask = data.mask | ~mask
    except Exception:
        # no mask
        pass

    return (data, (bounds, bounds_crs))


def render((bounds, bounds_crs),
           sources_store,
           shape,
           target_crs,
           format,
           transformation=None):
    """Render data intersecting bounds into shape using an optional
    transformation."""
    resolution_m = get_resolution_in_meters((bounds, bounds_crs), shape)
    stats = []

    if transformation:
        (bounds, bounds_crs), shape, offsets = transformation.expand(
            (bounds, bounds_crs), shape)

    with Timer() as t:
        sources = sources_store.get_sources((bounds, bounds_crs), resolution_m)
    stats.append(("get sources", t.elapsed))

    with Timer() as t:
        (sources_used, data, (data_bounds, data_crs)) = mosaic.composite(
            sources, (bounds, bounds_crs), shape, target_crs)
    stats.append(("composite", t.elapsed))

    if data is None:
        raise NoDataAvailable()

    data_format = "raw"

    if transformation:
        with Timer() as t:
            (data, data_format) = transformation.transform((data, (data_bounds,
                                                                   data_crs)))
        stats.append(("transform", t.elapsed))

        with Timer() as t:
            (data, (data_bounds, data_crs)) = transformation.postprocess(
                (data, (data_bounds, data_crs)), data_format, offsets)

        stats.append(("postprocess", t.elapsed))

    with Timer() as t:
        (content_type, formatted) = format((data, (data_bounds, data_crs)),
                                           data_format)
    stats.append(("format", t.elapsed))

    headers = {
        "Content-Type":
        content_type,
        "X-Imagery-Sources":
        ", ".join(s[1].split('/', 3)[3] for s in sources_used),
    }

    if os.environ.get('MARBLECUTTER_DEBUG_TIMERS'):
        headers.update({
            "X-Timers":
            ", ".join("{}: {:0.2f}".format(*s) for s in stats)
        })

    return (headers, formatted)



# coding=utf-8
from __future__ import absolute_import

import logging

from flask import Flask, jsonify

from flask_cors import CORS

from . import NoDataAvailable

LOG = logging.getLogger(__name__)

app = Flask("marblecutter")
CORS(app, send_wildcard=True)


class InvalidTileRequest(Exception):
    status_code = 404

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):  # noqa
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


@app.route("/favicon.ico")
def favicon():  # noqa
    return '', 404


@app.errorhandler(InvalidTileRequest)
def handle_invalid_tile_request(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code

    return response


@app.errorhandler(NoDataAvailable)
def handle_no_data_available(error):
    return "", 404


@app.errorhandler(IOError)
def handle_ioerror(error):
    LOG.warn(error)

    return "", 500



# coding=utf-8
from __future__ import absolute_import

from io import BytesIO

from PIL import Image

from .. import _isimage

CONTENT_TYPE = "image/png"


def PNG():
    def _format((data, (data_bounds, data_crs)), data_format):
        if not _isimage(data_format):
            raise Exception("Must be an image format")

        out = BytesIO()
        im = Image.fromarray(data, data_format.upper())
        im.save(out, "png")

        return (CONTENT_TYPE, out.getvalue())

    return _format



# noqa
# coding=utf-8
from __future__ import absolute_import, division, print_function

import logging

import numpy as np

from rasterio import warp

LOG = logging.getLogger(__name__)


def mask_outliers(data, m=2.):
    d = np.abs(data - np.median(data.compressed()))
    mdev = np.median(d.compressed())
    s = d / mdev if mdev else 0.
    return np.where(s < m, False, True)


def composite(sources, (bounds, bounds_crs), (height, width), target_crs):
    """Composite data from sources into a single raster covering bounds, but in
    the target CRS."""
    from . import _nodata, get_source, read_window

    ((left, right), (bottom, top)) = warp.transform(bounds_crs, target_crs,
                                                    bounds[::2], bounds[1::2])
    canvas = None
    canvas_bounds = (left, bottom, right, top)

    sources_used = list()

    # iterate over available sources, sorted by decreasing resolution
    for (url, source_name, resolution) in sources:
        with get_source(url) as src:
            if canvas is None:
                # infer the number of bands to use from the first available
                # source
                canvas = np.ma.zeros(
                    (src.count, height, width),
                    dtype=np.float32,
                    fill_value=_nodata(np.float32))
                canvas.mask = True
                canvas.fill_value = _nodata(np.float32)

            sources_used.append((source_name, url))

            LOG.info("Compositing %s (%s)...", url, source_name)

            # read a window from the source data
            # TODO ask for a buffer here, get back an updated bounding box
            # reflecting it
            # TODO NamedTuple for bounds (bounds + CRS)
            window_data = read_window(src, (canvas_bounds, target_crs),
                                      (height, width))

        if not window_data:
            continue

        data, _ = window_data
        if data.mask.any():
            # mask outliers (intended for DEM boundaries)
            LOG.info("masking outliers")
            data.mask[0] = np.logical_or(data.mask[0],
                                         mask_outliers(data[0], 100.))

        # paste (and reproject) the resulting data onto a canvas
        # TODO NamedTuple for data (data + bounds)
        canvas = paste(window_data, (canvas, (canvas_bounds, target_crs)))

        # TODO get the sub-array that contains nodata pixels and only fetch
        # sources that could potentially fill those (see
        # windows.get_data_window for the inverse)
        # See https://codereview.stackexchange.com/a/132933 for an example
        if not canvas.mask.any():
            # stop if all pixels are valid
            break

    return (sources_used, canvas, (canvas_bounds, target_crs))


def paste((window_data, (window_bounds, window_crs)), (canvas, (canvas_bounds,
                                                                canvas_crs))):
    """ "Reproject" src data into the correct position within a larger image"""
    if window_crs != canvas_crs:
        raise Exception(
            "CRSes must match: {} != {}".format(window_crs, canvas_crs))

    if window_bounds != canvas_bounds:
        raise Exception(
            "Bounds must match: {} != {}".format(window_bounds, canvas_bounds))

    if window_data.shape != canvas.shape:
        raise Exception("Data shapes must match: {} != {}".format(
            window_data.shape, canvas.shape))

    merged = np.ma.where(canvas.mask & ~window_data.mask, window_data, canvas)
    merged.fill_value = canvas.fill_value

    return merged



from setuptools import find_packages, setup

version = '0.2.3'

setup(
    name='marblecutter',
    version=version,
    description='Raster manipulation',
    url='https://github.com/mojodna/marblecutter',
    author='Seth Fitzsimmons',
    author_email='seth@mojodna.net',
    license='BSD',
    packages=find_packages(),
    zip_safe=False,
    package_data={
        'marblecutter': ['static/images/*', 'templates/*'],
    },
    install_requires=[
        'haversine',
        'mercantile',
        'numpy',
        'Pillow',
        'rasterio[s3]>=1.0a10',
        'scipy',
    ],
    extras_require={
        'color_ramp': 'matplotlib',
        'postgis': 'psycopg2',
        'web': [
            'flask',
            'flask-cors',
        ],
    },
)
